@@ -1,57 +1,40 @@
 /* -*- C++ -*-
+    This file contains a testsuite for job processing in ThreadWeaver.
 
-   This file contains a testsuite for job processing in ThreadWeaver.
+    SPDX-FileCopyrightText: 2005-2013 Mirko Boehm <mirko@kde.org>
 
-   $ Author: Mirko Boehm $
-   $ Copyright: (C) 2005-2013 Mirko Boehm $
-   $ Contact: mirko@kde.org
-         https://www.kde.org
-         http://creative-destruction.me $
-
-   This library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Library General Public
-   License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
-
-   This library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Library General Public License for more details.
-
-   You should have received a copy of the GNU Library General Public License
-   along with this library; see the file COPYING.LIB.  If not, write to
-   the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301, USA.
+    SPDX-License-Identifier: LGPL-2.0-or-later
 */
 
 #include "JobTests.h"
 
-#include <cstdlib>
 #include <algorithm>
+#include <cstdlib>
+#include <random>
 
 #include <QMutex>
+#include <QSignalSpy>
 #include <QTest>
 #include <QThread>
-#include <QSignalSpy>
 
-#include <ThreadWeaver/Queueing>
-#include <ThreadWeaver/QueueStream>
-#include <ThreadWeaver/Sequence>
-#include <ThreadWeaver/Lambda>
-#include <ThreadWeaver/ThreadWeaver>
-#include <ThreadWeaver/Thread>
-#include <ThreadWeaver/DebuggingAids>
 #include <ThreadWeaver/Collection>
-#include <ThreadWeaver/ResourceRestrictionPolicy>
+#include <ThreadWeaver/DebuggingAids>
 #include <ThreadWeaver/Dependency>
 #include <ThreadWeaver/DependencyPolicy>
-#include <ThreadWeaver/QObjectDecorator>
 #include <ThreadWeaver/Exception>
+#include <ThreadWeaver/Lambda>
+#include <ThreadWeaver/QObjectDecorator>
+#include <ThreadWeaver/QueueStream>
+#include <ThreadWeaver/Queueing>
+#include <ThreadWeaver/ResourceRestrictionPolicy>
+#include <ThreadWeaver/Sequence>
+#include <ThreadWeaver/Thread>
+#include <ThreadWeaver/ThreadWeaver>
 
-#include "AppendCharacterJob.h"
 #include "AppendCharacterAndVerifyJob.h"
-#include "WaitForIdleAndFinished.h"
+#include "AppendCharacterJob.h"
 #include "JobLoggingWeaver.h"
+#include "WaitForIdleAndFinished.h"
 
 QMutex s_GlobalMutex;
 
@@ -122,8 +105,8 @@ void JobTests::SimpleJobCollectionTest()
 {
     QString sequence;
     Collection jobCollection;
-    jobCollection << new AppendCharacterJob(QChar('a'), &sequence)
-                  << new AppendCharacterJob(QChar('b'), &sequence)
+    jobCollection << new AppendCharacterJob(QChar('a'), &sequence) //
+                  << new AppendCharacterJob(QChar('b'), &sequence) //
                   << new AppendCharacterJob(QChar('c'), &sequence);
 
     WaitForIdleAndFinished w(Queue::instance());
@@ -153,15 +136,15 @@ void JobTests::CollectionQueueingTest()
 {
     QString output;
     Collection jobCollection;
-    jobCollection << new AppendCharacterJob(QChar('a'), &output)
-                  << new AppendCharacterJob(QChar('b'), &output)
+    jobCollection << new AppendCharacterJob(QChar('a'), &output) //
+                  << new AppendCharacterJob(QChar('b'), &output) //
                   << new AppendCharacterJob(QChar('c'), &output);
 
     Queue weaver;
     WaitForIdleAndFinished w(&weaver);
     weaver.suspend();
     weaver.stream() << jobCollection;
-    QCOMPARE(weaver.queueLength(), 1); //collection queues itself, and its elements upon execution of self
+    QCOMPARE(weaver.queueLength(), 1); // collection queues itself, and its elements upon execution of self
     weaver.resume();
     weaver.finish();
     QCOMPARE(output.length(), 3);
@@ -177,10 +160,11 @@ QString SequenceTemplate = "abcdefghijklmnopqrstuvwxyz";
 class GeneratingCollection : public Collection
 {
 public:
-    void run(JobPointer, Thread *) override {
+    void run(JobPointer, Thread *) override
+    {
         std::for_each(SequenceTemplate.cbegin(), SequenceTemplate.cend(), [this](QChar it) {
             *this << new AppendCharacterJob(it, &sequence_);
-        } );
+        });
     }
     QString sequence_;
 };
@@ -188,10 +172,11 @@ public:
 class GeneratingSequence : public Sequence
 {
 public:
-    void run(JobPointer, Thread *) override {
+    void run(JobPointer, Thread *) override
+    {
         std::for_each(SequenceTemplate.cbegin(), SequenceTemplate.cend(), [this](QChar it) {
             *this << new AppendCharacterJob(it, &sequence_);
-        } );
+        });
     }
     QString sequence_;
 };
@@ -206,15 +191,15 @@ void JobTests::GeneratingCollectionTest()
     WaitForIdleAndFinished w(Queue::instance());
     stream() << collection;
     Queue::instance()->finish();
-    QCOMPARE(collection.sequence_.count(), SequenceTemplate.length());
+    QCOMPARE(collection.sequence_.size(), SequenceTemplate.length());
 }
 
 void JobTests::ShortJobSequenceTest()
 {
     QString sequence;
     Sequence jobSequence;
-    jobSequence << new AppendCharacterJob(QChar('a'), &sequence)
-                << new AppendCharacterJob(QChar('b'), &sequence)
+    jobSequence << new AppendCharacterJob(QChar('a'), &sequence) //
+                << new AppendCharacterJob(QChar('b'), &sequence) //
                 << new AppendCharacterJob(QChar('c'), &sequence);
 
     WaitForIdleAndFinished w(Queue::instance());
@@ -242,7 +227,8 @@ void JobTests::ShortDecoratedJobSequenceTest()
     jobSequence->addJob(jobB);
     jobSequence->addJob(jobC);
 
-    WaitForIdleAndFinished w(&queue); Q_UNUSED(w);
+    WaitForIdleAndFinished w(&queue);
+    Q_UNUSED(w);
     QVERIFY(DependencyPolicy::instance().isEmpty());
     queue.enqueue(jobSequence);
     // Job::DumpJobDependencies();
@@ -256,7 +242,8 @@ void JobTests::EmptyJobSequenceTest()
 {
     using namespace ThreadWeaver;
     QObjectDecorator sequence(new Sequence());
-    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+    WaitForIdleAndFinished w(Queue::instance());
+    Q_UNUSED(w);
     Q_ASSERT(Queue::instance()->isIdle());
     QSignalSpy doneSignalSpy(&sequence, SIGNAL(done(ThreadWeaver::JobPointer)));
     QCOMPARE(doneSignalSpy.count(), 0);
@@ -285,7 +272,7 @@ void JobTests::IncompleteCollectionTest()
 
     QString result;
     QObjectDecorator jobA(new AppendCharacterJob(QChar('a'), &result));
-    AppendCharacterJob jobB(QChar('b'), &result); //jobB does not get added to the sequence and queued
+    AppendCharacterJob jobB(QChar('b'), &result); // jobB does not get added to the sequence and queued
     QObjectDecorator col(new Collection());
     *col.collection() << jobA;
 
@@ -370,13 +357,14 @@ void JobTests::CollectionDependenciesTest()
     DependencyPolicy::instance().addDependency(Dependency(&col, jobC));
 
     // queue collection, but not jobC, the collection should not be executed
-    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+    WaitForIdleAndFinished w(Queue::instance());
+    Q_UNUSED(w);
     Queue::instance()->suspend();
     enqueue_raw(&col);
     Queue::instance()->resume();
     QCoreApplication::processEvents();
     QTest::qWait(100);
-    //FIXME verify: dfaure needed this here: QTRY_COMPARE(collectionStartedSignalSpy.count(), 0);
+    // FIXME verify: dfaure needed this here: QTRY_COMPARE(collectionStartedSignalSpy.count(), 0);
     QCOMPARE(collectionStartedSignalSpy.count(), 0);
     // enqueue jobC, first jobC then the collection should be executed
     Queue::instance()->enqueue(jobC);
@@ -384,8 +372,8 @@ void JobTests::CollectionDependenciesTest()
     Queue::instance()->finish();
     QVERIFY(col.isFinished());
     QVERIFY(result.startsWith(jobC->character()));
-    //QSKIP("This test is too fragile"); // PENDING(Mirko): fix
-    //QTRY_COMPARE(collectionStartedSignalSpy.count(), 1);
+    // QSKIP("This test is too fragile"); // PENDING(Mirko): fix
+    // QTRY_COMPARE(collectionStartedSignalSpy.count(), 1);
     loop.exec();
     qApp->processEvents();
     QCOMPARE(collectionStartedSignalSpy.count(), 1);
@@ -473,7 +461,7 @@ void JobTests::RecursiveSequenceTest()
     WaitForIdleAndFinished w(Queue::instance());
     Queue::instance()->suspend();
     Queue::instance()->enqueue(jobSequence4);
-//    DependencyPolicy::instance().dumpJobDependencies();
+    //    DependencyPolicy::instance().dumpJobDependencies();
     Queue::instance()->resume();
     Queue::instance()->finish();
     QCOMPARE(sequence, QLatin1String("abcdefghij"));
@@ -574,7 +562,7 @@ void JobTests::QueueAndDequeueAllCollectionTest()
     Queue::instance()->suspend();
     QVERIFY(Queue::instance()->isEmpty());
     Queue::instance()->enqueue(collection);
-    //collection cannot have been started, so only one job is queued at the moment:
+    // collection cannot have been started, so only one job is queued at the moment:
     QCOMPARE(Queue::instance()->queueLength(), 1);
     Queue::instance()->dequeue();
     QVERIFY(Queue::instance()->isEmpty());
@@ -638,7 +626,6 @@ void JobTests::RecursiveQueueAndDequeueAllCollectionTest()
     QVERIFY(Queue::instance()->isEmpty());
     Queue::instance()->resume();
     Queue::instance()->finish();
-
 }
 
 void JobTests::RecursiveQueueAndDequeueAllSequenceTest()
@@ -727,8 +714,8 @@ void JobTests::SimpleRecursiveSequencesTest()
     jobSequence1 << new AppendCharacterJob(QChar('b'), &sequence);
 
     Sequence jobSequence2;
-    jobSequence2 << new AppendCharacterJob(QChar('a'), &sequence)
-                 << jobSequence1
+    jobSequence2 << new AppendCharacterJob(QChar('a'), &sequence) //
+                 << jobSequence1 //
                  << new AppendCharacterJob(QChar('c'), &sequence);
 
     WaitForIdleAndFinished w(Queue::instance());
@@ -801,8 +788,8 @@ void JobTests::ResourceRestrictionPolicyBasicsTest()
     AppendCharacterJob g('g', &sequence);
     Collection collection;
     collection << a << b << c << d << e << f << g;
-    const QList<AppendCharacterJob*> lstJob = { &a, &b, &c, &d, &e, &f, &g };
-    for (AppendCharacterJob* job : lstJob) {
+    const QList<AppendCharacterJob *> lstJob = {&a, &b, &c, &d, &e, &f, &g};
+    for (AppendCharacterJob *job : lstJob) {
         QMutexLocker l(job->mutex());
         job->assignQueuePolicy(&restriction);
     }
@@ -827,7 +814,7 @@ void JobTests::JobSignalsAreEmittedAsynchronouslyTest()
 {
     using namespace ThreadWeaver;
 
-    char bits[] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g' };
+    char bits[] = {'a', 'b', 'c', 'd', 'e', 'f', 'g'};
     const int NumberOfBits = sizeof bits / sizeof bits[0];
     QString sequence;
     QObjectDecorator collection(new Collection, this);
@@ -855,18 +842,20 @@ void JobTests::deliveryTestJobDone(JobPointer)
     deliveryTestCounter.fetchAndAddRelease(-1);
 }
 
-void noOp() {}
+void noOp()
+{
+}
 
 void JobTests::JobSignalsDeliveryTest()
 {
-    //This test was added to investigate segmentation faults during signal delivery from jobs to the main thread.
-    //Relies on processEvents() processing all pending events, as the specification says.
+    // This test was added to investigate segmentation faults during signal delivery from jobs to the main thread.
+    // Relies on processEvents() processing all pending events, as the specification says.
     using namespace ThreadWeaver;
 
     QCOMPARE(deliveryTestCounter.loadAcquire(), 0);
     WaitForIdleAndFinished w(Queue::instance());
     for (int count = 0; count < 100; ++count) {
-        QJobPointer job(new QObjectDecorator(new Lambda<void(*)()>(noOp)));
+        QJobPointer job(new QObjectDecorator(new Lambda<void (*)()>(noOp)));
         QVERIFY(connect(job.data(), SIGNAL(done(ThreadWeaver::JobPointer)), SLOT(deliveryTestJobDone(ThreadWeaver::JobPointer))));
         deliveryTestCounter.fetchAndAddRelease(1);
         Queue::instance()->enqueue(job);
@@ -884,14 +873,14 @@ void decrementCounter()
 
 void JobTests::JobPointerExecutionTest()
 {
-    //This test was added to investigate segmentation faults during signal delivery from jobs to the main thread.
-    //Relies on processEvents() processing all pending events, as the specification says.
+    // This test was added to investigate segmentation faults during signal delivery from jobs to the main thread.
+    // Relies on processEvents() processing all pending events, as the specification says.
     using namespace ThreadWeaver;
 
     QCOMPARE(deliveryTestCounter.loadAcquire(), 0);
     WaitForIdleAndFinished w(Queue::instance());
     for (int count = 0; count < 100; ++count) {
-        JobPointer job(new Lambda<void(*)()>(decrementCounter));
+        JobPointer job(new Lambda<void (*)()>(decrementCounter));
         deliveryTestCounter.fetchAndAddRelease(1);
         Queue::instance()->enqueue(job);
     }
@@ -943,7 +932,11 @@ void JobTests::IdDecoratorSingleAllocationTest()
     struct DecoratedJob : public IdDecorator {
         QString sequence;
         AppendCharacterJob job;
-        DecoratedJob() : IdDecorator(&job, false), job('a', &sequence) {}
+        DecoratedJob()
+            : IdDecorator(&job, false)
+            , job('a', &sequence)
+        {
+        }
     };
 
     WaitForIdleAndFinished w(Queue::instance());
@@ -965,7 +958,7 @@ struct InstanceCountedJob : public Job {
         counter.fetchAndAddRelease(1);
     }
 
-    ~InstanceCountedJob()
+    ~InstanceCountedJob() override
     {
         counter.fetchAndAddRelease(-1);
     }
@@ -981,7 +974,8 @@ QAtomicInt InstanceCountedJob::counter;
 void JobTests::JobsAreDestroyedAfterFinishTest()
 {
     using namespace ThreadWeaver;
-    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+    WaitForIdleAndFinished w(Queue::instance());
+    Q_UNUSED(w);
     Queue::instance()->suspend();
     JobPointer job(new InstanceCountedJob);
     Queue::instance()->enqueue(job);
@@ -1002,7 +996,8 @@ void JobTests::JobExitStatusByExceptionTest()
     using namespace ThreadWeaver;
 
     struct FailingJob : public Job {
-        void run(JobPointer, Thread *) override {
+        void run(JobPointer, Thread *) override
+        {
             throw JobFailed();
         }
     };
@@ -1012,7 +1007,8 @@ void JobTests::JobExitStatusByExceptionTest()
     QCOMPARE(failing.status(), Job::Status_Failed);
 
     struct AbortingJob : public Job {
-        void run(JobPointer, Thread *) override {
+        void run(JobPointer, Thread *) override
+        {
             throw JobAborted();
         }
     };
@@ -1022,7 +1018,8 @@ void JobTests::JobExitStatusByExceptionTest()
     QCOMPARE(aborting.status(), Job::Status_Aborted);
 
     struct SuccessfulJob : public Job {
-        void run(JobPointer, Thread *) override {
+        void run(JobPointer, Thread *) override
+        {
             // do nothing
         }
     };
@@ -1036,7 +1033,8 @@ void JobTests::JobManualExitStatusTest()
     using namespace ThreadWeaver;
 
     struct FailingJob : public Job {
-        void run(JobPointer, Thread *) override {
+        void run(JobPointer, Thread *) override
+        {
             setStatus(Job::Status_Failed);
         }
     };
@@ -1046,7 +1044,8 @@ void JobTests::JobManualExitStatusTest()
     QCOMPARE(failing.status(), Job::Status_Failed);
 
     struct AbortingJob : public Job {
-        void run(JobPointer, Thread *) override {
+        void run(JobPointer, Thread *) override
+        {
             setStatus(Job::Status_Aborted);
         }
     };
@@ -1056,7 +1055,8 @@ void JobTests::JobManualExitStatusTest()
     QCOMPARE(aborting.status(), Job::Status_Aborted);
 
     struct SuccessfulJob : public Job {
-        void run(JobPointer, Thread *) override {
+        void run(JobPointer, Thread *) override
+        {
             // do nothing
         }
     };
@@ -1069,31 +1069,36 @@ void JobTests::QueueStreamLifecycleTest()
 {
     QString sequence;
     using namespace ThreadWeaver;
-    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+    WaitForIdleAndFinished w(Queue::instance());
+    Q_UNUSED(w);
     stream() << make_job(new AppendCharacterJob('a', &sequence)) // enqueues JobPointer
              << new AppendCharacterJob('b', &sequence) // enqueues JobInterface*
              << make_job(new AppendCharacterJob('c', &sequence));
     Queue::instance()->finish();
-    QCOMPARE(sequence.count(), 3);
+    QCOMPARE(sequence.size(), 3);
 }
 
-class SynchronizedNumbers {
+class SynchronizedNumbers
+{
 public:
-    void append(int number) {
+    void append(int number)
+    {
         QMutexLocker l(&mutex_);
         numbers_.append(number);
     }
 
-    bool isSorted() const {
+    bool isSorted() const
+    {
         QMutexLocker l(&mutex_);
         return std::is_sorted(numbers_.cbegin(), numbers_.cend());
     }
 
-    void sortChunks(int chunkSize) {
+    void sortChunks(int chunkSize)
+    {
         QMutexLocker l(&mutex_);
         Q_ASSERT(numbers_.count() % chunkSize == 0);
         auto start = numbers_.begin();
-        while(start!=numbers_.end()) {
+        while (start != numbers_.end()) {
             auto stop = start;
             std::advance(stop, chunkSize);
             std::sort(start, stop);
@@ -1102,19 +1107,24 @@ public:
     }
 
 private:
-    QVector<int> numbers_;
+    QList<int> numbers_;
     mutable QMutex mutex_;
 };
 
-class GeneratingEnumeratorSequence : public ThreadWeaver::Sequence {
+class GeneratingEnumeratorSequence : public ThreadWeaver::Sequence
+{
 public:
-    GeneratingEnumeratorSequence(SynchronizedNumbers* numbers, int start, int count)
-        : start_(start), count_(count), numbers_(numbers)
-    {}
+    GeneratingEnumeratorSequence(SynchronizedNumbers *numbers, int start, int count)
+        : start_(start)
+        , count_(count)
+        , numbers_(numbers)
+    {
+    }
 
-    void run(JobPointer, Thread*) override {
+    void run(JobPointer, Thread *) override
+    {
         numbers_->append(start_);
-        for(int index = start_ + 1; index < start_+count_; ++index) {
+        for (int index = start_ + 1; index < start_ + count_; ++index) {
             *this << new GeneratingEnumeratorSequence(numbers_, index, 1);
         }
     }
@@ -1122,43 +1132,51 @@ public:
 private:
     const int start_;
     const int count_;
-    SynchronizedNumbers* numbers_;
+    SynchronizedNumbers *numbers_;
 };
 
-class GeneratingEnumeratorCollection : public ThreadWeaver::Collection {
+class GeneratingEnumeratorCollection : public ThreadWeaver::Collection
+{
 public:
-    GeneratingEnumeratorCollection(SynchronizedNumbers* numbers, int start, int count)
-        : start_(start), count_(count), numbers_(numbers)
-    {}
+    GeneratingEnumeratorCollection(SynchronizedNumbers *numbers, int start, int count)
+        : start_(start)
+        , count_(count)
+        , numbers_(numbers)
+    {
+    }
 
-    void run(JobPointer, Thread*) override {
+    void run(JobPointer, Thread *) override
+    {
         numbers_->append(start_);
-        QVector<GeneratingEnumeratorCollection*> elements;
-        for(int index = start_ + 1; index < start_+count_; ++index) {
+        QList<GeneratingEnumeratorCollection *> elements;
+        for (int index = start_ + 1; index < start_ + count_; ++index) {
             elements.append(new GeneratingEnumeratorCollection(numbers_, index, 1));
         }
-        std::random_shuffle(elements.begin(), elements.end());
-        std::for_each(elements.begin(), elements.end(), [this](QVector<GeneratingEnumeratorCollection>::iterator it) {
-            *this << *it;
-        } );
+        std::random_device rd;
+        std::mt19937 g(rd());
+        std::shuffle(elements.begin(), elements.end(), g);
+        std::for_each(elements.begin(), elements.end(), [this](GeneratingEnumeratorCollection *collection) {
+            *this << collection;
+        });
     }
 
 private:
     const int start_;
     const int count_;
-    SynchronizedNumbers* numbers_;
+    SynchronizedNumbers *numbers_;
 };
 
 void JobTests::NestedGeneratingCollectionsTest()
 {
     using namespace ThreadWeaver;
-    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+    WaitForIdleAndFinished w(Queue::instance());
+    Q_UNUSED(w);
 
     SynchronizedNumbers numbers;
     const int NumberOfSequences = 100;
     const int ElementsPerSequence = 20;
     Sequence sequence;
-    for(int index = 0; index < NumberOfSequences; ++index) {
+    for (int index = 0; index < NumberOfSequences; ++index) {
         sequence << new GeneratingEnumeratorCollection(&numbers, index * ElementsPerSequence, ElementsPerSequence);
     }
     stream() << sequence;
@@ -1167,15 +1185,17 @@ void JobTests::NestedGeneratingCollectionsTest()
     QVERIFY(numbers.isSorted());
 }
 
-void JobTests::NestedGeneratingSequencesTest() {
+void JobTests::NestedGeneratingSequencesTest()
+{
     using namespace ThreadWeaver;
-    WaitForIdleAndFinished w(Queue::instance()); Q_UNUSED(w);
+    WaitForIdleAndFinished w(Queue::instance());
+    Q_UNUSED(w);
 
     SynchronizedNumbers numbers;
     const int NumberOfSequences = 100;
     const int ElementsPerSequence = 20;
     Sequence sequence;
-    for(int index = 0; index < NumberOfSequences; ++index) {
+    for (int index = 0; index < NumberOfSequences; ++index) {
         sequence << new GeneratingEnumeratorSequence(&numbers, index * ElementsPerSequence, ElementsPerSequence);
     }
     stream() << sequence;
@@ -1189,14 +1209,16 @@ void JobTests::DeeperNestedGeneratingCollectionsTest()
     using namespace ThreadWeaver;
     auto logger = new JobLoggingWeaver();
     Queue queue(logger);
-    WaitForIdleAndFinished w(&queue); Q_UNUSED(w);
+    WaitForIdleAndFinished w(&queue);
+    Q_UNUSED(w);
     const int ElementsPerCollection = 20;
     const int NumberOfBlocks = 2;
     const int CollectionsPerBlock = 2;
-    SynchronizedNumbers numbers; Sequence sequence;
-    for(int block=0; block < NumberOfBlocks; ++block) {
+    SynchronizedNumbers numbers;
+    Sequence sequence;
+    for (int block = 0; block < NumberOfBlocks; ++block) {
         auto col = new Collection();
-        for(int collection = 0; collection < CollectionsPerBlock; ++collection) {
+        for (int collection = 0; collection < CollectionsPerBlock; ++collection) {
             const int start = (block * NumberOfBlocks + collection) * ElementsPerCollection;
             *col << new GeneratingEnumeratorCollection(&numbers, start, ElementsPerCollection);
         }
@@ -1204,11 +1226,203 @@ void JobTests::DeeperNestedGeneratingCollectionsTest()
     }
     queue.stream() << sequence;
     queue.finish();
-    numbers.sortChunks(NumberOfBlocks*ElementsPerCollection);
+    numbers.sortChunks(NumberOfBlocks * ElementsPerCollection);
     QVERIFY(numbers.isSorted());
 }
 
+struct AbortableJob : public Job {
+    int aborted = 0;
+    int extraCode = 0;
+    QMutex waitForStart;
+    QMutex waitForAbort;
 
+    AbortableJob()
+    {
+        waitForStart.lock();
+        waitForAbort.lock();
+    }
+
+    ~AbortableJob()
+    {
+        waitForStart.unlock();
+        waitForAbort.unlock();
+    }
+
+    void run(JobPointer, Thread *) override
+    {
+        QCOMPARE(this->status(), Job::Status_Running);
+        waitForStart.unlock();
+        waitForAbort.lock();
+        if (shouldAbort()) {
+            aborted++;
+            throw JobAborted();
+        }
+        extraCode++;
+    }
+
+    void requestAbort() override
+    {
+        Job::requestAbort();
+
+        // Let it handle abortion
+        waitForAbort.unlock();
+    }
+};
+
+void JobTests::RequestAbortCollectionTest()
+{
+    AbortableJob abortable;
+
+    QString sequence;
+    Collection jobCollection;
+    jobCollection << new AppendCharacterJob(QChar('a'), &sequence) << abortable << new AppendCharacterJob(QChar('c'), &sequence);
+
+    WaitForIdleAndFinished w(Queue::instance());
+    stream() << jobCollection;
+
+    // We need to wait for job to have started execution
+    abortable.waitForStart.lock();
+
+    // Now tell it to abort
+    jobCollection.requestAbort();
+
+    w.finish();
+
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+    QVERIFY(sequence.length() == 2);
+    QVERIFY(sequence.count('a') == 1);
+    QVERIFY(sequence.count('c') == 1);
+    QVERIFY(abortable.aborted == 1);
+    QVERIFY(abortable.extraCode == 0);
+    QCOMPARE(jobCollection.status(), Job::Status_Aborted);
+    QVERIFY(abortable.waitForStart.tryLock() == false);
+    QVERIFY(abortable.waitForAbort.tryLock() == false);
+}
+
+void JobTests::RequestAbortSequenceTest()
+{
+    AbortableJob abortable;
+
+    QString sequence;
+    Sequence jobSequence;
+    jobSequence << new AppendCharacterJob(QChar('a'), &sequence) << abortable << new AppendCharacterJob(QChar('c'), &sequence);
+
+    WaitForIdleAndFinished w(Queue::instance());
+    stream() << jobSequence;
+
+    // We need to wait for job to have started execution
+    abortable.waitForStart.lock();
+
+    // Now tell it to abort
+    jobSequence.requestAbort();
+
+    w.finish();
+
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+    QCOMPARE(sequence, QLatin1String("a"));
+    QVERIFY(abortable.aborted == 1);
+    QVERIFY(abortable.extraCode == 0);
+    QCOMPARE(jobSequence.status(), Job::Status_Aborted);
+    QVERIFY(abortable.waitForStart.tryLock() == false);
+    QVERIFY(abortable.waitForAbort.tryLock() == false);
+}
+
+void JobTests::JobOnFinishTest()
+{
+    QString sequence;
+    int finishedA = 0;
+    int finishedB = 0;
+    Collection jobCollection;
+    jobCollection << new AppendCharacterJob(QChar('a'), &sequence) << new AppendCharacterJob(QChar('b'), &sequence)
+                  << new AppendCharacterJob(QChar('c'), &sequence);
+
+    jobCollection.onFinish([&](const JobInterface &job) {
+        QCOMPARE(job.status(), Job::Status_Success);
+
+        // Result should already be correct now
+        QVERIFY(sequence.length() == 3);
+        QVERIFY(sequence.count('a') == 1);
+        QVERIFY(sequence.count('b') == 1);
+        QVERIFY(sequence.count('c') == 1);
+
+        // We should get same original instance
+        QVERIFY(&job == &jobCollection);
+
+        finishedA++;
+    });
+
+    jobCollection.onFinish([&](const JobInterface &job) {
+        QCOMPARE(jobCollection.status(), Job::Status_Success);
+
+        QVERIFY(sequence.length() == 3);
+        // We should get same original instance
+        QVERIFY(&job == &jobCollection);
+
+        finishedB++;
+    });
+
+    WaitForIdleAndFinished w(Queue::instance());
+    stream() << jobCollection;
+    w.finish();
+
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+    QVERIFY(finishedA == 1);
+    QVERIFY(finishedB == 1);
+}
+
+void JobTests::JobOnFinishAbortTest()
+{
+    QString sequence;
+    int finishedA = 0;
+    int finishedB = 0;
+    Sequence jobSequence;
+
+    struct AbortingJob : public Job {
+        void run(JobPointer, Thread *) override
+        {
+            throw JobAborted();
+        }
+    };
+
+    AbortingJob aborting;
+
+    jobSequence << new AppendCharacterJob(QChar('a'), &sequence) << aborting << new AppendCharacterJob(QChar('c'), &sequence);
+
+    jobSequence.onFinish([&](const JobInterface &job) {
+        // Result should already be correct now
+        QCOMPARE(sequence, QLatin1String("a"));
+
+        // We should get same original instance
+        QVERIFY(&job == &jobSequence);
+
+        // We should be in aborted state
+        QCOMPARE(job.status(), Job::Status_Aborted);
+
+        finishedA++;
+    });
+
+    jobSequence.onFinish([&](const JobInterface &job) {
+        // Result should already be correct now
+        QCOMPARE(sequence, QLatin1String("a"));
+
+        // We should get same original instance
+        QVERIFY(&job == &jobSequence);
+
+        // We should be in aborted state
+        QCOMPARE(jobSequence.status(), Job::Status_Aborted);
+
+        finishedB++;
+    });
+
+    WaitForIdleAndFinished w(Queue::instance());
+    stream() << jobSequence;
+    w.finish();
+
+    QVERIFY(DependencyPolicy::instance().isEmpty());
+    QVERIFY(finishedA == 1);
+    QVERIFY(finishedB == 1);
+}
 
 QTEST_MAIN(JobTests)
 
+#include "moc_JobTests.cpp"
